import matplotlib.pyplot as plt

# PyTorch
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.distributions import Categorical

# Gym
import gym
import pygame


env_id = "CartPole-v1"
# Create the env
env = gym.make(env_id, render_mode="human")

# Get the state space and action space
s_size = env.observation_space.shape[0]
a_size = env.action_space.n

# 4 entrées : la position du cart, sa vitesse, l'angle de la barre, sa vitesse angulaire 
# 2 sorties : aller à droite et aller a gauche
agent = nn.Sequential(nn.Linear(s_size, 128),
                         nn.ReLU(),
                         nn.Dropout(),
                         nn.Linear(128,a_size),
                         nn.Softmax()) 

print(agent)

optimizer = optim.Adam(agent.parameters(), lr=5e-3)

episode_rewards = []
gamma=0.99

for _ in range (200):
    observation = env.reset()
    
    rewards = []
    log_probs = []
    terminated = False
    total_reward = 0

    while not(terminated):
        # Convert observation to tensor
        observation_array = observation[0] if isinstance(observation, tuple) else observation
        observation_tensor = torch.FloatTensor(observation_array).unsqueeze(0)
        # observation_tensor = torch.FloatTensor(observation).unsqueeze(0)
        # Compute action probabilities
        probs = agent(observation_tensor)
        m = Categorical(probs)
        # Sample action
        action = m.sample()
        log_prob = m.log_prob(action)
        observation, reward, terminated, truncated, info = env.step(action.item())
        env.render()
        
        log_probs.append(log_prob)
        rewards.append(reward)
        total_reward += reward
        

    episode_rewards.append(total_reward)

    # Compute returns
    R = 0
    returns = []
    for r in rewards[::-1]:
        R = r + gamma * R
        returns.insert(0, R)
    returns = torch.tensor(returns)
    returns = (returns - returns.mean()) / (returns.std() + 1e-5)

    # Compute policy loss
    policy_loss = []
    for log_prob, R in zip(log_probs, returns):
        policy_loss.append(log_prob * R)
    policy_loss = - torch.cat(policy_loss).sum()

    # Update policy
    optimizer.zero_grad()
    policy_loss.backward()
    optimizer.step()

    print(f'Episode {_} finished')

env.close()


plt.plot(episode_rewards)
plt.title('Total Reward per Episode')
plt.xlabel('Episode')
plt.ylabel('Total Reward')
plt.show()